#include <algorithm>
#include <cstddef>
#include <iterator>
#include <limits>
#include <scc/memory.h>
#include <vector>

template <unsigned long long SIZE, unsigned BUSWIDTH> class flash : public scc::memory<SIZE, scc::LT> {

public:
    flash(const sc_core::sc_module_name& nm);
    cci::cci_param<unsigned> FB_BUFFER_COUNT{"FB_BUFFER_COUNT", 2, "number of flash buffers"};
    cci::cci_param<unsigned> FB_SIZE{"FB_SIZE", 2, "represents the flash line size normalized to the bus data width"};
    cci::cci_param<unsigned> FR_DELAY{
        "FR_DELAY", 0,
        "When a read is requested to the flash, it takes FR_DELAY cycles longer to respond in case of a buffer miss"};
    cci::cci_param<sc_core::sc_time> FR_DELAY_TIME{"FR_DELAY_TIME", sc_core::SC_ZERO_TIME,
                                                   "When a read is requested to the flash, it takes FR_DELAY_TIME "
                                                   "duration longer to respond in case of a buffer miss"};

private:
    std::vector<unsigned> buffers;
    std::vector<unsigned> timestamps;
    unsigned counter{0};
    int handle_operation(tlm::tlm_generic_payload& trans, sc_core::sc_time& delay);
    bool buffer_hit(uint64_t addr);
};

template <unsigned long long SIZE, unsigned BUSWIDTH>
flash<SIZE, BUSWIDTH>::flash(const sc_core::sc_module_name& nm)
: scc::memory<SIZE, scc::LT>(nm)
, buffers(FB_BUFFER_COUNT, std::numeric_limits<unsigned>::max())
, timestamps(FB_BUFFER_COUNT, 0) {
    this->set_operation_callback(
        [this](scc::memory<SIZE, scc::LT>& mem, tlm::tlm_generic_payload& gp, sc_core::sc_time& delay) -> int {
            tlm::tlm_command cmd = gp.get_command();
            auto res = 0;
            if(cmd == tlm::TLM_READ_COMMAND) {
                ::sc_dt::uint64 adr = gp.get_address();
                unsigned len = gp.get_data_length();
                delay += buffer_hit(adr)               ? sc_core::SC_ZERO_TIME
                         : this->clk_i.get_interface() ? this->clk_i->read() * FR_DELAY
                                                       : FR_DELAY_TIME;
                res = mem.handle_operation(gp, delay);
            } else if(!sc_core::sc_time_stamp().value()) {
                res = mem.handle_operation(gp, delay);
            }
            gp.set_dmi_allowed(false);
            return res;
        });
}
template <unsigned long long SIZE, unsigned BUSWIDTH> bool flash<SIZE, BUSWIDTH>::buffer_hit(uint64_t addr) {
    auto current_cache_start = addr - addr % (BUSWIDTH / 8 * FB_SIZE);
    auto it = std::find(buffers.begin(), buffers.end(), current_cache_start);
    if(it != buffers.end()) {
        size_t index = std::distance(buffers.begin(), it);
        timestamps[index] = ++counter;
        return true;
    } else {
        auto it = std::min_element(timestamps.begin(), timestamps.end());
        size_t lru_index = std::distance(timestamps.begin(), it);
        buffers[lru_index] = current_cache_start;
        timestamps[lru_index] = ++counter;
        return false;
    }
}